/*
 *  TODO:
 *    * version control on db, through comment header 
 *
 *   <pyramidsignature>
 *   <p v='0.12'     rgb='0.1 0.1 0.1'>
 *   \t<p v='0.1'    rgb='0.1 0.1 0.1'/>
 *   \t<p v='0.11'   rgb='0.2 0.3 0.2'>
 *   \t\t<p v='0.14' rgb='0.1 0.2 0.2'/>
 *   \t\t<p v=''/>
 *   \t\t<p v=''/>
 *   \t\t<p v=''/>
 *   \t</p>
 *   \t<p v='' rgb_m=''>
 *   </p>
 *   </pyramidsignature>
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "db.h"
#include "image.h"
#include "digest.h"

#define INBUF_LEN 320000

struct _akois_db
{
  int                 image_count;
  struct akois_entry *images;
};

static inline struct akois_entry *
akois_entries_add (struct akois_entry *self,
                   struct akois_entry *entry)
{
  struct akois_entry *iter;
  if (!self)
    return entry;
  iter=self;
  while (iter->next)
    iter=iter->next;
  iter->next =entry;
  return self;
}


akois_db *
akois_db_new (void)
{
  akois_db *self;

  self = calloc (1, sizeof (struct _akois_db));
  self->image_count = 0;

  return self;
}

void
akois_db_free (akois_db *self)
{
  struct akois_entry *entry=self->images;
  while (entry)
    {
      struct akois_entry *temp=entry;
      if (temp->image)
	akois_image_free (temp->image); 
      entry=temp->next;
      free (temp);
    }
  free (self);
}

void
akois_db_load (akois_db *self,
               const char *file)
{
  FILE *fp;
  int i=0;
  char buf[INBUF_LEN];

  fp = fopen (file, "r");
  if (!fp)
    return;
 
  fprintf (stdout, "\rloading database %s         \n", file);
  while (fgets (buf, INBUF_LEN, fp))
    {
      if (buf[0]=='#'   ||
          buf[0]=='\0'  ||
          buf[0]=='\n')
        {
          /* comment line read */
        }
      else
        {
          struct akois_entry *entry = calloc (sizeof (struct akois_entry), 1);

          entry->image = akois_image_new_from_str (buf);
          if (entry->image)
            {
              self->images = akois_entries_add (self->images, entry);
              self->image_count ++;
              i++;
            }
          else
            {
              free (entry);
            }
        }

      fprintf (stdout, "\r%c (%i records)", i%4==0?'|':
                                            i%4==1?'/':
                                            i%4==2?'-':'\\', i);
      fflush(0);
    }
  fclose (fp);
  fprintf (stdout, "\n"); 
  
  return;
}

void
akois_db_save (akois_db   *self,
               const char *path)
{
  FILE *file;
  struct akois_entry *iter = self->images;
  int i = 0;
  fprintf (stdout, "\rsaving database to %s         \n", path);
  file = fopen (path, "w");
  fprintf (file, "#lines starting with # are comments\n");
  fprintf (file, "#as are blank lines, this file contains\n");
  fprintf (file, "#a data of the form %s,\n", akois_digest_get_type ());
  fprintf (file, "\n");
  fprintf (file, "#<file_path> <digest>\n");
  fprintf (file, "\n");
  fprintf (file, "#written by akois\n");
  fprintf (file, "\n");

  while (iter)
    {
      char *img_buf = akois_image_to_str (iter->image);
      if (img_buf)
        {
         fprintf (file, "%s\n", img_buf);
         free (img_buf);
	}
      iter = iter->next;

      fprintf (stdout, "\r%c %2.2f%% (%i/%i)", i%4==0?'|':
               i%4==1?'/':
               i%4==2?'-':'\\', (100.0*(i+1))/self->image_count,
              (i+1), self->image_count);
      i++;
      fflush(0);
    }
  fclose (file);
}

static int
akois_db_contains_image (akois_db   *self,
                         const char *path)
{
  struct akois_entry *iter = self->images;
  while (iter)
    {
      if (!strcmp (path, akois_image_get_path (iter->image)))
        return 1;
      iter = iter->next;
    }
  return 0;
}



/* note, ownership is passed to db,. and db will
 * free the image on shutdown
 */  
void
akois_db_add_image (akois_db    *self,
                    akois_image *image)
{
  struct akois_entry *entry;

  if (!image || akois_db_contains_image (self, akois_image_get_path (image)))
    {
      return;
    }

  entry  = calloc (sizeof (struct akois_entry), 1);
  entry->image=image;

  akois_image_calc_digest (image);

  self->images = akois_entries_add (self->images, entry);
  self->image_count++;
}


void
akois_db_add_image_path (akois_db   *self,
                         const char *image_path)
{
  akois_image *image;

  if (akois_db_contains_image (self, image_path))
    {
      /*fprintf (stderr, "%s already present; ignoring.\n", image_path);*/
      return;
    }


  image = akois_image_new ();
  akois_image_set_path (image, image_path);

  akois_db_add_image (self, image);
}

void          
akois_db_best_many (akois_db       *self,
                    akois_options  *options,
                    akois_image   **samples,
                    akois_image   **results,
                    int             n
                   )
{
  float     best_distance[n];
  int       i;

  for (i=0; i<n; i++)
    if (samples[i]!=NULL)
    {
      best_distance[i]= 4242424242424242.0;
      results[i] = NULL;
    }

  struct akois_entry *iter = self->images;
  
  while (iter)
    {
      for (i=0; i<n; i++)
        if (samples[i]!=NULL)
          {
            float distance;
        
            distance = akois_image_distance (samples[i], iter->image, options);
            if (distance < best_distance[i])
              {
                best_distance[i] = distance;
                results[i]= iter->image;
              }
          }

      akois_image_inc_compares (iter->image);
      iter = iter->next;
    }

  for (i=0; i<n; i++)
    if (samples[i]!=NULL)
      akois_image_inc_hits (results[i]);
}

akois_image *
akois_db_best (akois_db      *self,
               akois_image   *sample,
               akois_options *options)
{
  float    best_distance = 4242424242424242.0;
  int       temporal_shadow = akois_options_get_temporal_shadow (options);

  akois_image *best_image = sample; /* initially include the sample, as the
                                       worst match,. this might help in the
                                       first period when the db is smaller than
                                       temporal shadow. */
  struct akois_entry *iter = self->images;
  int                i     = 0;

  
  while (iter && i< self->image_count &&
         !(self->image_count < temporal_shadow + i))
    {
      float distance;
      
      distance = akois_image_distance (sample, iter->image, options);
      if (distance < best_distance)
        {
          best_distance = distance;
          best_image    = iter->image;
        }

      akois_image_set_compares (iter->image, akois_image_get_compares (iter->image)+1);
      iter = iter->next;
      i++;
    }

  akois_image_set_hits (best_image, akois_image_get_hits (best_image)+1);
  return best_image;
}

static struct akois_entry *
simple_db_query (struct akois_entry   *images,
                 akois_image   *sample,
                 akois_options *options,
                 int            image_count)
{
  struct akois_entry *iter = images;
  struct akois_entry *first = NULL;
  int                 i     = 0;

  while (iter && i<image_count)
    {
      struct akois_entry *item;

      if (image_count < akois_options_get_temporal_shadow (options) + i)
        { 
          /* image is amongst the last few to be ignored
           *
           * if implemented properly as a temporal shadow,. images after and
           * before can be ignored, thus allowing better tuning for the domain.
           */

          iter = iter->next;
          i++;
          continue;
        }
      
      item = malloc (sizeof( struct akois_entry));
      item->image = iter->image;

      item->distance = akois_image_distance (sample, item->image, options);

      if (!first)
        {
          first = item;
          first->next = NULL;
        }
      else if (item->distance < first->distance)
        {
          item->next = first;
          first = item;
        }
      else
        {
          /* pay attention to scoping rules!!! similarily named
           * variables, should probably be broken down into separate
           * functions
           */
          struct akois_entry *iter=first;
          while (iter->next && item->distance > iter->next->distance)
            iter=iter->next;
          item->next = iter->next;
          iter->next = item;
        }
      iter = iter->next;
      i++;
    }

  akois_image_freshen (sample);
  return first;
}


struct akois_entry *
akois_db_query (akois_db      *self,
                akois_image   *sample,
                akois_options *options)
{
  struct akois_entry *first = NULL;
  int neighbourhood = akois_options_get_neighbourhood (options);


  first = simple_db_query (self->images, sample, options, self->image_count);

  if (neighbourhood>0)
    {
      struct akois_entry *iter = first;
      struct akois_entry *revised_results = NULL;
      int i=0;
      /* assign all neighbours a rank */
      while (iter && i<neighbourhood)
        {
          int ranked=0;
          struct akois_entry *secondary;
          struct akois_entry *secondary_iter;

          struct akois_entry *item;
          item = malloc (sizeof( struct akois_entry));
          item->image = iter->image;

          secondary = simple_db_query (first, iter->image, options, neighbourhood); 
          /* FIXME: the query should only be done amongst the neighborhood
           * and not on the entire database,. thus either the first list
           * should be chopped,. or a max number of compared should be passed
           * in, the latter is least destructive for a "reordering",..
           * but the images in the end returned will not have the inversion
           * in place...
           */

          secondary_iter = secondary;
          while (secondary_iter && (secondary_iter->image != sample))
            {
              secondary_iter = secondary_iter->next;
              ranked++;
            }
          akois_entries_free (secondary);

          item->rank     = ranked;
          item->image    = iter->image;
          item->distance = iter->distance;

          /* insertion */

          if (!revised_results)
            {
              revised_results = item;
              revised_results->next = NULL;
            }
          else if (item->rank < revised_results->rank)
            {
              item->next = revised_results;
              revised_results = item;
            }
          else
            {
              /* pay attention to scoping rules!!! similarily named
               * variables, should probably be broken down into separate
               * functions
               */
              struct akois_entry *iter=revised_results;
              while (iter->next && item->rank > iter->next->rank)
                iter=iter->next;
              item->next = iter->next;
              iter->next = item;
            }
          i++;

          iter=iter->next;
        }
      /* after the ranks have been defined,. it needs to be sorted,..
       * a better solution might be to create a linked list, and insert
       * the items incrementally
       */
      akois_entries_free (first);
      return revised_results;
      
    }
  
  return first;
}

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <glob.h>
#include <libgen.h>

void
akois_db_add_images (akois_db   *self,
                     const char *filter)
{
  glob_t globbuf;
  int i;

  fprintf (stdout, "\radding images %s\n", filter);
  glob (filter, GLOB_PERIOD | GLOB_MARK, NULL, &globbuf);
  /*for (i=0; i<1 && i<globbuf.gl_pathc;i++)*/
  for (i=0; i<globbuf.gl_pathc; i++)
  {
   fprintf (stdout, "\r%c %2.2f%% (%i/%i) %s      ", i%4==0?'|':
                                              i%4==1?'/':
                                              i%4==2?'-':'\\', (100.0*(i+1))/globbuf.gl_pathc,
                                              (i+1), (int)globbuf.gl_pathc, globbuf.gl_pathv[i]);
   fflush(0);
    akois_db_add_image_path (self, globbuf.gl_pathv[i]);
  }
  fprintf (stdout, "\n");
  globfree (&globbuf);
}


akois_image *
akois_db_image (akois_db *self,
                int       no)
{
  int i=0;
  struct akois_entry *iter = self->images;
  while (iter)
    {
      if (i==no)
        return iter->image;
      i++;
      iter = iter->next;
    }
  return self->images->image;
}

int
akois_db_count (akois_db *self)
{
  return self->image_count;
}

/*****************/
                          
void
akois_db_zero_hits (akois_db *self)
{
  struct akois_entry *iter = self->images;
  while (iter)
    {
      akois_image_set_hits (iter->image, 0);
      iter = iter->next;
    }
}

static int static_slack = 100;

/*          |sacred|  |slack|
 */
static int hits_cmp(const void *a,
                     const void *b)
{
  long scoreA;
  long scoreB;
  long cmps;
  const struct akois_entry *e_a = *(void **)a;
  const struct akois_entry *e_b = *(void **)b;

  cmps = akois_image_get_compares (e_a->image);
  if (cmps==0)
    scoreA = 1000000;
  else if (cmps< static_slack)
    scoreA = 100000000 * akois_image_get_hits (e_a->image) / cmps;
  else
    scoreA = 1000000 * akois_image_get_hits (e_a->image) / cmps; 

  cmps = akois_image_get_compares (e_b->image);
  if (cmps==0)
    scoreB = 1000000;
  else if (cmps< static_slack)
    scoreB = 100000000 * akois_image_get_hits (e_b->image) / cmps;
  else
    scoreB = 1000000 * akois_image_get_hits (e_b->image) / cmps; 

  return scoreB-scoreA;
}

static int first=1;
static void evict (akois_image *image)
{
          /* move image to eviction directory */
          {
            char *eviction_dir="/munin-evicted/";
            char *temp_path = strdup (akois_image_get_path (image));
            char *temp_path2 = strdup (temp_path);
            char *base = basename (temp_path2);
            char new_path[512];
            char *base2 = strdup (base);
            base2[16]='\0';

            if (first)
              {
                first=0;
                sprintf (new_path, "/%s/%s", eviction_dir, base2);
                mkdir (new_path, 511);
              }

            sprintf (new_path, "/%s/%s/%s", eviction_dir, base2, base);
            if(rename (temp_path, new_path))
              {
                if (unlink (temp_path))
                  fprintf (stderr, "rename and remove failed");
              }

            free (base2);
            free (temp_path);
            free (temp_path2);
          }
          akois_image_free (image);
}

#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

/* reduce the database size to new_size, based on hits,
 * slack is the amount of extra images allowed without action
 * taken (used to implement database caches)
 */
void
akois_db_reduce_to (akois_db *self,
                    int       new_size,
                    int       slack)
{
  struct akois_entry  *iter = self->images;
  struct akois_entry **temp_array;
  int i;
  int evict_first=1;
 
 static_slack = slack; 
  if (self->image_count < new_size + slack)
    return;

  temp_array = malloc (sizeof(struct akois_entry*) * self->image_count);
  
  i=0;
  while (iter)
    {
      temp_array[i]=iter;
      iter = iter->next;
      i++;
    }

  qsort (temp_array, self->image_count, sizeof (struct akois_entry*), hits_cmp);

  self->images = NULL;

  if (self->image_count - 1 >= new_size)
    evict_first = 0;
  
  i=0;
  if (evict_first && temp_array[i]->image)
    { 
      evict (temp_array[i]->image);
      free (temp_array[i]);
      i++;
    }
  for (; i< new_size + evict_first; i++)
    {
      struct akois_entry *entry = temp_array[i];
      entry->next = self->images;
      self->images = entry;
    }
  for (; i< self->image_count; i++)
    {
      if (temp_array[i]->image)
        {
          evict (temp_array[i]->image);
          free (temp_array[i]);
        }
    }

  self->image_count = new_size;
  fprintf (stdout, "reduced to %i\n", new_size);
  free (temp_array);
  first=1; /* reset the first image helper variable to true for next run */
}
