/** a database of images */

#ifndef _AKOIS_DB_H
#define _AKOIS_DB_H

#include "image.h"

typedef struct _akois_db     akois_db;

struct akois_entry {
  float               distance;   /* the signature space distance to the query image */
  int                 rank;       /* when doing neighbourhood reversal, the rank distance
                                     to original query image */

  akois_image        *image;
  struct akois_entry *next;
};

/* akois query results is returned as a linked list */

akois_db *
akois_db_new (void);

void
akois_db_free (akois_db *self);

void
akois_db_load (akois_db   *self,
               const char *file);

void
akois_db_save (akois_db   *self,
               const char *file);

void
akois_db_add_image (akois_db    *self,
                    akois_image *image);

void
akois_db_add_image_path (akois_db   *self,
                         const char *image_path);

struct akois_entry *
akois_db_query (akois_db      *self,
                akois_image   *sample,
                akois_options *options);

char **
akois_db_query_by_name (akois_db    *self,
                        const char  *sample_name,
                        int          depth);

void
akois_db_add_images (akois_db   *self,
                     const char *filter);

/* optimized version returning only the single best match,
 * the need arose from the deja vu mirror project, which
 * wants tight code paths for speed reasons
 */
akois_image *
akois_db_best (akois_db      *self,
               akois_image   *sample,
               akois_options *options);


/* query the best matches for n images simultanously, if an image
 * in the query is set to NULL it is ignored. The resulting images
 * are stored in the passed in results array
 */
void          
akois_db_best_many (akois_db       *self,
                    akois_options  *options,
                    akois_image   **samples,
                    akois_image   **results,
                    int             n
                   );

akois_image *
akois_db_image (akois_db *self,
                int       no);

int
akois_db_count (akois_db *self);

void
akois_db_zero_hits (akois_db *self);

void
akois_db_reduce_to (akois_db *self,
                    int       new_size,
                    int       slack);

static inline void
akois_entries_free (struct akois_entry *self)
{
  while (self)
    {
      struct akois_entry *temp=self;
      self=self->next;
      free (temp);
    }
}


#endif /* _AKOIS_DB_H */
