#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "digest.h"


#define HORIZONTAL 24
#define VERTICAL   24

#define overlap 0.8

#define OUTBUF_LEN  (VERTICAL * HORIZONTAL * 3 * 10 + 256)

struct _akois_digest
{
  int           digested;
  float         vector[HORIZONTAL*VERTICAL * 3];
};

void
dejab_digest_freshen (akois_digest *self);

enum data_item
{
  data_complexity,
  data_mean_y,
  data_mean_u,
  data_mean_v,
  data_count
};

akois_digest *
dejab_digest_new (void)
{
  akois_digest *self;

  self = calloc (1, sizeof(struct _akois_digest));
  self->digested = 0;
  return self;
}

void
dejab_digest_free (akois_digest *self)
{
  dejab_digest_freshen (self);
  free (self);
}

static void
extract_rgb_mean (int width,
                  int height,
                  int row_stride,
                  unsigned char *pixels,
                  int x0, int y0,
                  int x1, int y1,
                  float *mean_r,
                  float *mean_g,
                  float *mean_b)
  {
  int sample_count = 0;
  long   t_mean_r = 0;
  long   t_mean_g = 0;
  long   t_mean_b = 0;
    int x,y;

    for (y=y0;y<y1;y++)
      for (x=x0;x<x1;x++)
        {
          if ( (x>=0 && x<width) &&
               (y>=0 && y<height))
            {
              t_mean_r += pixels[y*row_stride+x*3+0];
              t_mean_g += pixels[y*row_stride+x*3+1];
              t_mean_b += pixels[y*row_stride+x*3+2];
              sample_count ++;
            }
        }

    *mean_r = t_mean_r / (sample_count * 255.0);
    *mean_g = t_mean_g / (sample_count * 255.0);
    *mean_b = t_mean_b / (sample_count * 255.0);
}


void
dejab_digest_calc (akois_digest  *self,

                   unsigned char *pixels,
                   int            width,
                   int            height,
                   int            row_stride,

                   int            x0_px,
                   int            y0_px,
                   int            width_px,
                   int            height_px)
{
  int x,y;

  for (y=0;y<VERTICAL;y++)
    for (x=0;x<HORIZONTAL;x++)
      {
        float red, green, blue;
        int x0 = (x0_px  +((float)(x-overlap)/HORIZONTAL)*width_px);
        int y0 = (y0_px  +((float)(y-overlap)/VERTICAL)*height_px);
        int x1 = (x0_px  +((float)(x+1+overlap)/HORIZONTAL)*width_px);
        int y1 = (y0_px  +((float)(y+1+overlap)/VERTICAL)*height_px);
        if (x0==x1) x1++;
        if (y0==y1) y1++;

        extract_rgb_mean (width, height, row_stride, pixels,
         x0,y0,x1,y1,

                  &red, &green, &blue);

        self->vector [(y*HORIZONTAL + x) * 3 + 0] = red;
        self->vector [(y*HORIZONTAL + x) * 3 + 1] = green;
        self->vector [(y*HORIZONTAL + x) * 3 + 2] = blue;
      }

  self->digested = 1;
}

char *
dejab_digest_to_str (akois_digest *self)
{
  int i;
  char buf[OUTBUF_LEN];
  char *s= &buf[0];
  if (!self->digested)
      return NULL;
  for (i=0; i<HORIZONTAL * VERTICAL * 3; i++)
    {
      s+=sprintf (s, "%2.4f ", self->vector[i]);
    }
  
  return strdup(buf);
}

akois_digest *
dejab_digest_new_from_str (const char *str)
{
  int i;
  char *s= (char*)str;
  akois_digest *self;
  self = dejab_digest_new ();

  for (i=0;i<HORIZONTAL*VERTICAL*3;i++)
    {
      char *ts = NULL;
      self->vector[i] = strtod (s, &ts);
      s=ts;
    }
  self->digested = 1;
  return self;
}

double
dejab_digest_compare (akois_digest  *self,
                      akois_digest  *other,
                      akois_options *options)
{
  int length;
  int j;
  float distance = 0.0;

  length = HORIZONTAL * VERTICAL;

  for (j=0;j<length;j++)
    {
      int i;
      float sum=0.0;
      for (i=0;i<3;i++)
        {
          float component = (self->vector[j*3+i] - other->vector[j*3+i]);
          component *= component;
          switch (i)
            {
              case 0:
                component *= 0.7;
                break;
              case 1:
                component *= 1.0;
                break;
              case 2:
                component *= 0.7;
                break;
            }
          sum+=component;
        }
      //sum = sqrt (sum);
      distance += sum;
    }

  /* Add in the global average match as well*/
  {
     double sum=0;
     double avg[2][3]={{0,0,0}, {0,0,0}};

     for (j=0;j<length;j++)
     {
	 int i;
	 for (i=0;i<3;i++)
          avg[0][i] += self->vector[j*3+i];
	 for (i=0;i<3;i++)
          avg[1][i] += other->vector[j*3+i];
     }
     for (j=0;j<3;j++)
     {
	 float component = avg[0][j]-avg[1][j];
	 component/=length;
	 component *= component;
	 sum+=component;
     }
     distance += sum;
  }

  distance = distance / (double)(HORIZONTAL*VERTICAL+1);

  return distance;
}


void
dejab_digest_freshen (akois_digest *self)
{
}

void
dejab_digest_write_png (akois_digest *self, const char *path){};

