/* FIXME: write out hits statistics */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libgen.h>
#include <math.h>
#include <babl/babl.h>

#include "image.h"
#include "options.h"
#include "digest.h"
#include "load_image.h"

#define OUTBUF_LEN 320000

struct _akois_image
{
  char          *path;
  akois_digest  *digest;

  akois_digest  *sub_digest;
  unsigned char *buf;
  int            w;
  int            h;
  int            stride;
  int            hits;        /* used by mosaic and deja vu databases */
  int            compares;    /* used by mosaic and deja vu databases */
};

akois_image *
akois_image_new (void)
{
  akois_image *self;

  self = calloc (1, sizeof(struct _akois_image));
  self->path       = strdup ("");
  self->digest     = NULL;
  self->sub_digest = NULL;
  self->buf        = NULL;
  self->w          = 0;
  self->h          = 0;
  self->stride     = 0;
  self->hits      = 0;
  self->compares  = 0;
  return self;
}

void
akois_image_free (akois_image *self)
{
  akois_image_freshen (self);
  free (self->path);
  if (self->digest)
    akois_digest_free (self->digest);
  if (self->sub_digest)
    akois_digest_free (self->sub_digest);
  free (self);
}

#include "load_image.h"
int
akois_image_get_data (akois_image    *self,
                      int            *width,
                      int            *height,
                      int            *row_stride,
                      unsigned char **pixels)
{
 if (self->buf==NULL)
   {
      unsigned char *buf = NULL;
      int w, h, stride;
      if (!self->path ||
          !strcmp(self->path, ""))
        return -1;
      if (!load_image (self->path, &w, &h, &stride, &buf))
        {
          self->buf = buf;
          self->stride = stride;
          self->w = w;
          self->h = h;
        }
   }
 if (self->buf!=NULL)
   {
     *width = self->w;
     *height = self->h;
     *row_stride = self->stride;
     *pixels = self->buf;
     return 0;
   }
 return -1;
}

int
akois_image_get_hits (akois_image *self)
{
  return self->hits;
}

void 
akois_image_set_hits (akois_image *self,
                       int          hits)
{
  self->hits=hits;
}

void
akois_image_inc_hits (akois_image *self)
{
  self->hits++;
}

void
akois_image_inc_compares (akois_image *self)
{
  self->compares++;
}

int
akois_image_get_compares (akois_image *self)
{
  return self->compares;
}

void 
akois_image_set_compares (akois_image *self,
                           int          compares)
{
  self->compares=compares;
}


void
akois_image_freshen (akois_image *self)
{
  if (self->digest)
    akois_digest_freshen (self->digest);
  if (self->sub_digest)
    akois_digest_free (self->sub_digest);
  self->sub_digest = NULL;
  if (self->buf)
    free (self->buf);
  self->buf = NULL;
}

void
akois_image_set_path (akois_image *self,
                      const char  *path)
{
  free (self->path);
  self->path = strdup (path);
}

const char *
akois_image_get_path (akois_image *self)
{
  return self->path;
}

void
akois_image_write_png (akois_image *self,
                       const char  *path)
{
  if (self->digest)
    akois_digest_write_png (self->digest, path);
}

void
akois_image_calc_digest (akois_image *self)
{
  akois_digest *digest;

  if (self->digest)
    return; /* oh the horror, one bug that was found was an: 
               an akois_digest_free (self->digest); before the 
               return,. nastyness*/

  digest = akois_digest_new ();
    {
       int w, h, stride;
       unsigned char *buf = NULL;

       if (!akois_image_get_data (self, &w, &h, &stride, &buf))
         {
           akois_digest_calc (digest,
                              buf, w, h, stride, 
                              0, 0, w, h);
         }
    }
  self->digest = digest;
}

/* An additional query image feature, originally developed
 * for subimage search.
 */
void
akois_image_calc_sub_digest (akois_image *self,
                             double       x0,
                             double       y0,
                             double       width,  
                             double       height)
{ 
  akois_digest *digest;

  if (self->sub_digest)
    akois_digest_free (self->sub_digest);

  digest = akois_digest_new ();
    {
       if (!self->buf)
         {
            unsigned char *buf = NULL;
            int w, h, stride;
            load_image (self->path, &w, &h, &stride, &buf);
  //          downsize_to (512, &w, &h, &stride, &buf);
            self->buf = buf;
            self->stride = stride;
            self->w = w;
            self->h = h;
         }

       if (self->buf)
         {
           akois_digest_calc (digest,
                              self->buf, self->w, self->h, self->stride, 
                              x0 * self->w, y0 * self->h,
                              width * self->w, height * self->h);
           //free (buf);
         }
    }
  self->sub_digest = digest;
}

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

akois_image *
akois_image_new_from_str (const char *str)
{
  akois_image *self = akois_image_new ();
  char         tmp[512];
  const char  *s=str;
  char        *d=tmp;
  int          hits;
  int          compares;

  while(*s && *s!=' ')
    *d++ = *s++;
  *d='\0';
  if (*s==' ')
    s++;

  hits = atoi (s);

  while(*s && *s!=' ')
    s++;
  if (*s==' ')
    s++;

  compares = atoi (s);

  while(*s && *s!=' ')
    s++;
  if (*s==' ')
    s++;
 
  {
    struct stat buf;
    if (stat (tmp, &buf))
      {
        free (self);
        return NULL;
      }
  } 
  akois_image_set_path (self, tmp);
  self->digest = akois_digest_new_from_str (s);
  self->hits = hits;
  self->compares = compares;

  return self;
}

static char buf[OUTBUF_LEN];
char *
akois_image_to_str (akois_image *self)
{
  char *digest = akois_digest_to_str (self->digest);
  if (digest)
  {
     sprintf (buf, "%s %i %i %s", self->path, self->hits, self->compares, digest);
     free (digest);
     return strdup(buf);
  }
  return NULL;
}

double
akois_image_distance (akois_image   *self,
                      akois_image   *other,
                      akois_options *options)
{
  if (!self || !other)
    return 100000;
  if (!self->digest ||
      !other->digest)
    {
       fprintf (stderr, "self->digest: %p other->digest: %p\n\n",
                (void *)self->digest, (void *)other->digest);
       return 100000.0;
    }
  
  if (self->sub_digest)
    return akois_digest_compare (self->sub_digest,
                                 other->digest,
                                 options);
  return akois_digest_compare (self->digest,
                               other->digest,
                               options);
}

void
akois_image_adopt_sub_digest (akois_image *self,
                              akois_image *other)
{
  if (self->digest)
    akois_digest_free (self->digest);
  self->digest=NULL;

  if (!other->sub_digest)
    {
      fprintf (stderr, "%s:%i %s(), no subdigest\n",
       __FILE__, __LINE__, __FUNCTION__);
    }
  self->digest = other->sub_digest;
  other->sub_digest = NULL;
}
