#include "dispatch.h"
#include "lost-in-time.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <SDL.h>

static double     threshold      = 1.0;
static int        USE_FULLSCREEN = 0;

static void parse_args (int    argc, char **argv);

static int  oxiderun_iteration ();
static void quadtree_init ();
static void mirror_init ();
static void ghost_init ();
static void splash_init ();
static void oxiderun_destroy ();

static int  event_iteration ();
static int running=1;

int
munin_splash (int argc,
              char **argv)
{
  Dispatch *dispatch;

  parse_args (argc, argv);
  dispatch = dispatch_new ();

  dispatch_add_task (dispatch, 0, splash_init, oxiderun_iteration, oxiderun_destroy, NULL);
  dispatch_add_task (dispatch, 0, NULL, event_iteration, NULL, NULL);

  running = 1;
  while (dispatch_iterate (dispatch) && running);

  dispatch_destroy (dispatch);
  return 0;
}

int
munin_quadtree (int argc,
                char **argv)
{
  Dispatch *dispatch;

  parse_args (argc, argv);
  dispatch = dispatch_new ();

  dispatch_add_task (dispatch, 0, quadtree_init, oxiderun_iteration, oxiderun_destroy, NULL);
  dispatch_add_task (dispatch, 0, NULL, event_iteration, NULL, NULL);

  running = 1;
  while (dispatch_iterate (dispatch) && running);

  dispatch_destroy (dispatch);
  return 0;
}

int
munin_mirror (int argc,
              char **argv)
{
  Dispatch *dispatch;

  parse_args (argc, argv);
  dispatch = dispatch_new ();

  dispatch_add_task (dispatch, 0, mirror_init, oxiderun_iteration, oxiderun_destroy, NULL);
  dispatch_add_task (dispatch, 0, NULL, event_iteration, NULL, NULL);

  running = 1;
  while (dispatch_iterate (dispatch) && running);

  dispatch_destroy (dispatch);
  return 0;
}

int
munin_ghost  (int argc,
              char **argv)
{
  Dispatch *dispatch;

  parse_args (argc, argv);
  dispatch = dispatch_new ();

  dispatch_add_task (dispatch, 0, ghost_init, oxiderun_iteration, oxiderun_destroy, NULL);
  dispatch_add_task (dispatch, 0, NULL, event_iteration, NULL, NULL);

  running = 1;
  while (dispatch_iterate (dispatch) && running);

  dispatch_destroy (dispatch);
  return 0;
}

static int
event_iteration ()
{
  SDL_Event event;
  while (SDL_PollEvent (&event))
   {
     switch (event.type)
       {
         case SDL_QUIT:
           vm_set_task ("quit");
           running = 0;
           break;
         case SDL_KEYUP:
           switch (event.key.keysym.sym)
             {
               default:
                 running = vm_key_handle (event.key.keysym.sym);
                 break;
             }
           break;
         default:
           break;
       }
   }
  return 1;
}


#include <oxide/oxide.h>
static Oxide *oxide = NULL;

static int CAPTURE_WIDTH=128;
static int CAPTURE_HEIGHT=96;

#include <gggl/image/gggl_image.h>

#include <math.h>

int oxiderun_iteration ()
{
  oxide_process (oxide);
  return 1;
}


static void
parse_args (int    argc,
            char **argv)
{
  char **targ = argv;
  
  targ++; /* ignore program name*/ 
  while (*targ)
    {
      if (!strcmp (*targ, "-t"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-t expects argument\n");
            }
          threshold = atof (*targ);
        }
      else if (!strcmp (*targ, "-w"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-w expects argument\n");
            }
          CAPTURE_WIDTH = atoi (*targ);
        }
      else if (!strcmp (*targ, "-h"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-h expects argument\n");
            }
          CAPTURE_HEIGHT = atoi (*targ);
        }
      else if (!strcmp (*targ, "-h"))
        {
          targ++;
          USE_FULLSCREEN = 1;
        }
      targ++;
    }
}

/* eye */

#include <gggl/gggl.h>          /* for loading additional op dir */

/* global variables are bad, but an easy way to implement
 * command line arguments
 */

#define DEBUG_MODE
#define MBLUR_STR_EXHIBIT "0.01"
#define MBLUR_STR_MOTION "0.6"

#define   MBLUR_STR MBLUR_STR_MOTION

static void
quadtree_init (void)
{
   oxide = oxide_parse_xml_chunk (NULL, "<oxide>"
      "  <op type='display_yuv' title='quadtree'/>"
      "  <op type='hflip' />"
      "  <op type='over'>"
      "     <op type='opacity'  level='0.5'/>"
      "     <op type='quadtree' level='0.7'/>"
      "     <op type='clone'    idref='5' />"
      "  </op>"
      "  <op type='quadtree'    level='0.5'/>"
     /* "  <op type='forcefmt' pixfmt='g8'>"
      "    <output id='5'/>"
      "  </op>"*/
      "  <op type='v4l'>"
      "    <property name='width'>640</property>"
      "    <property name='height'>480</property>"
      "  </op>"
      "</oxide>");
}

static void
mirror_init (void)
{
   oxide = oxide_parse_xml_chunk (NULL, "<oxide>"
      "  <op type='display_yuv' title='mirror'/>"
      "  <op id='Source' type='v4l'>"
      "    <property name='width'>640</property>"
      "    <property name='height'>480</property>"
      "  </op>"
      "</oxide>");
}

static void
ghost_init (void)
{
   oxide = oxide_parse_xml_chunk (NULL, 
    "<oxide>"
    "<op type='display_yuv' title='ghost'/>"
    "<op type='over'>"
    "    <property name='x'>0</property>"
    "    <property name='y'>0</property>"
    "    <property name='subpixel_positioning'>1</property>"
    "    <op type='apply_mask'>"
    "        <op type='median'>"
    "            <property name='horizontal'>1</property>"
    "            <property name='vertical'>1</property>"
    "        </op>"
    "        <op type='invert'>"
    "        </op>"
    "        <op type='key_clean_slate'>"
    "            <property name='level'>0.03</property>"
    "            <property name='smooth_range'>0.03</property>"
    "            <op name='clone of boxblur' type='clone'>"
    "                <property name='idref'>16</property>"
    "            </op>"
    "        </op>"
    "        <op name='mblur' type='mblur'>"
    "            <property name='opacity'>0.09</property>"  /* amount of reveal */
    "            <output id='8'/>"
    "        </op>"
    "        <op name='boxblur' type='boxblur'>"
    "            <property name='horizontal'>2</property>"
    "            <property name='vertical'>2</property>"
    "            <output id='16'/>"
    "        </op>"
    "        <op name='clone of webcam' type='clone'>"
    "            <property name='idref'>17</property>"
    "        </op>"
    "    </op>"
    "    <op name='clone of webcam' type='clone'>"
    "        <property name='idref'>17</property>"
    "    </op>"
    "  </op>"
    "<op name='mblur' type='mblur'>"
    "    <property name='opacity'>0.0001</property>"   /* background update*/
    "</op>"
    "<op name='hflip' type='hflip'>"
    "    <output id='17'/>"
    "  </op>"
    "<op name='webcam' type='v4l'>"
    "<!--"
    "  <property name='width'>160</property>"
    "  <property name='height'>120</property>"
    "  -->"
    "  <property name='width'>160</property>"
    "  <property name='height'>120</property>"
    "  </op>"
    "</oxide> ");
}

static void
splash_init (void)
{
   oxide = oxide_parse_xml_chunk (NULL, "<oxide>"
      "  <op type='display-sdl' title='void'/>"
      "  <op id='Source' type='blank'>"
      "    <property name='width'>640</property>"
      "    <property name='height'>480</property>"
      "  </op>"
      "</oxide>");
}

static void
oxiderun_destroy ()
{
  oxide_free (oxide);
}
